"""
This module manages pyrrot's configuration.
"""

from enum import auto, Enum
from os import path, environ
import configparser
import json

class SelectionMode(Enum):
    """
    SelectionMode represents the different selection choices.
    """
    ALL = auto()
    SELECTION = auto()
    ALBUMART = auto()

class WallpaperConfig():
    """
    This class represents the configuration of pyrrot.
    It handles the various accessible options.
    """

    def __init__(self, config_file=None) -> None:
        self.config_file = config_file or \
            path.join(
                environ.get('XDG_CONFIG_HOME') or
                environ.get('APPDATA') or
                path.join(environ['HOME'], '.config'),
                "pyrrot/pyrrot.config"
            )
        print(self.config_file)
        if not path.exists(config_file):
            raise FileNotFoundError(config_file)
        config = configparser.ConfigParser()
        config.read(config_file)

        self.debug = config["global"]["debug"] == "true"
        selection_mode_possibilities = {
            "all": SelectionMode.ALL,
            "selection": SelectionMode.SELECTION,
            "albumart": SelectionMode.ALBUMART
        }
        if config["global"]["mode"] not in ("all", "selection", "albumart"):
            raise Exception("Error: Mode is wrong ! Must be one of all, selection or albumart.")

        self.selection_mode = selection_mode_possibilities[config["global"]["mode"]]

        self.metadata_file = config["global"]["picture_infos"]

        self.music = {
            "directory": config["music"]["music_dir"],
            "albumart_path": config["music"]["albumart_path"]
        }

        self.selection_options = {
            "include_tags":     [tag.strip() for tag in config["selection_options"]["include_tags"].strip("[]").split(',')],
            "include_colours":  [colour.strip() for colour in config["selection_options"]["include_colours"].strip("[]").split(',')],
            "include_files":    [filename.strip() for filename in config["selection_options"]["include_files"].strip("[]").split(',')],
            "exclude_tags":     [tag.strip() for tag in config["selection_options"]["exclude_tags"].strip("[]").split(',')],
            "exclude_colours":  [colour.strip() for colour in config["selection_options"]["exclude_colours"].strip("[]").split(',')],
            "exclude_files":    [filename.strip() for filename in config["selection_options"]["exclude_files"].strip("[]").split(',')]
        }

        for option_name, option_value in self.selection_options.items():
            if "" in option_value:
                self.selection_options[option_name].remove("")

        if self.debug:
            print(json.dumps(self.selection_options, sort_keys=True, indent=2))

        self.theme = {
            "do_update_theme": config["theme"]["update_theme"] == "true",
            "use_static_theme": config["theme"]["use_static_theme"] == "true",
            "default_theme": config["theme"]["default_theme"],
            "powerline_colours": config["theme"]["powerline_colours"]
        }



"""
This module handles wallpaper metadata.
"""

import json
from os.path import exists
from typing import Dict, List, Tuple
import random
from pyrrot_wallpaper.config import SelectionMode, WallpaperConfig


class WallpaperMetadata():
    """
    This class manages Wallpaper metadata
    """

    def __init__(self, image_directory, metdata_file) -> None:
        """
        :param string metadata_file: Path to the metadata json file
        """
        self.image_directory = image_directory
        with open(metdata_file, 'r', encoding="utf-8") as file:
            extracted_metadata = json.loads(file.read())
            self.metadata = extracted_metadata
            metadata_validity = self.is_infofile_valid()
            if not metadata_validity[0]:
                raise Exception(metadata_validity[1])

    def get_wallpapers_with_colours(self, colours: List[str]) -> List[Dict]:
        """
        :param List[str] colours: List of the colours to get
        :return: List of wallpapers with at least a colour in the list colours.
        :rtype: List[Dict]
        """
        res = []
        for pic in self.metadata:
            for colour in pic["colours"]:
                if colour in colours:
                    res.append(pic)
                    break
        return res

    def get_wallpapers_with_tags(self, tags: List[str]) -> List[Dict]:
        """
        :param List[str] tags: List of the tags to get
        :return: List of wallpapers with at least a tag in the list tags.
        :rtype: List[Dict]
        """

        res = []
        for pic in self.metadata:
            for tag in pic["tags"]:
                if tag in tags:
                    res.append(pic)
                    break
        return res

    def is_infofile_valid(self) -> Tuple[bool, str]:
        """The file has been opened with json lib, so we do not need to check
        whether it is a valid json file.
        Two checks are made :
        The first one is that the picture has a name.
        The second one is whether the file truly exists.

        :param dict infos: Metadata about the collection of wallpapers
        :return: (True, "") if the file is valid, else (False, "<Error message>")
        :rtype: Tuple(bool, str)
        """
        if not(isinstance(self.metadata, list)) or len(self.metadata) == 0:
            return False, "Error: The info file does not contain any image !"
        for picture in self.metadata:
            if not "name" in picture:
                return False, f"Error: Picture {str(picture)} has no name !"
            if "file" not in picture:
                return False, f"Error: Picture \"{picture['name']}\": no \"file\" is defined !"
            imagefile = self.image_directory + "/" + picture["file"]
            if not exists(imagefile):
                return False, \
                f"Error: Picture \"{picture['name']}\": file {imagefile} does not exist"
        return True, ""

    def wallpaper_selection_list(self, wallpaper_config: WallpaperConfig) -> List:
        """
        :param selection_mode: Selection mode
        :param wallpaper_config: Wallpaper configuration
        :return: A list of the usable dictionaries
        :rtype: List
        """
        selected_pictures = []

        if wallpaper_config.selection_mode != SelectionMode.SELECTION:
            return self.metadata

        for pic in self.metadata:
            excluded = False
            for tag in wallpaper_config.selection_options["exclude_tags"]:
                if tag in pic["tags"]:
                    excluded = True
                    break
            for colour in wallpaper_config.selection_options["exclude_colours"]:
                if colour in pic["colours"]:
                    excluded = True
                    break
            for file in wallpaper_config.selection_options["exclude_files"]:
                if file in pic["file"]:
                    excluded = True
                    break
            if not excluded:
                if len(wallpaper_config.selection_options["include_tags"]) == 0:
                    selected_pictures.append(pic)
                else:
                    for tag in pic["tags"]:
                        if tag in wallpaper_config.selection_options["include_tags"] \
                            and pic not in selected_pictures:
                            selected_pictures.append(pic)
                            break
                if len(wallpaper_config.selection_options["include_colours"]) > 0:
                    selected_pictures.append(pic)
                else:
                    for colour in pic["colours"]:
                        if colour in wallpaper_config.selection_options["include_colours"] \
                            and pic not in selected_pictures:
                            selected_pictures.append(pic)
                            break
                if len(wallpaper_config.selection_options["include_files"]) > 0:
                    selected_pictures.append(pic)
                else:
                    if pic["file"] in wallpaper_config.selection_options["include_files"] \
                        and pic not in selected_pictures:
                        selected_pictures.append(pic)
        return selected_pictures

    def select_single_wallpaper(self, wallpaper_config: WallpaperConfig, random_seed=None) -> Dict:
        """
        :param WallpaperConfig wallpaper_config:
        :param (int|None) random_seed: Seed for the random generator.
        :return: A single picture, which abides by the selection criteria.
        :rtype: Dict
        """
        random.seed(random_seed)
        selected_pictures = self.wallpaper_selection_list(wallpaper_config)
        if len(selected_pictures) == 0:
            raise Exception("There is no wallpaper matching your criteria.")
        pic = random.choice(selected_pictures)
        print(json.dumps(pic, sort_keys=True, indent=4))
        return pic



"""
This module handles the setting of a wallpaper.
"""

from os.path import dirname, expanduser
import subprocess
import pywal

from pyrrot_wallpaper.config import SelectionMode,WallpaperConfig

class WallpaperSetter():
    """
    This class manages the setting of a wallpaper.
    """

    def __init__(self) -> None:
        pass

    def get_wallpaper_full_path(self, wallpaper_dict: dict, wallpaper_config: WallpaperConfig) \
         -> str:
        """
        :param Dict wallpaper_dict: Wallpaper dictionary, with the "file" key either non-existent
        or containing the path relative to the metadata file
        : param WallpaperConfig wallpaper_config: Pyrrot configuration
        :return: Full path to the image
        :rtype: str
        """
        if wallpaper_config.selection_mode == SelectionMode.ALBUMART:
            fullpath = expanduser(wallpaper_config.music["albumart_path"])
        else:
            fullpath = expanduser(dirname(wallpaper_config.metadata_file)) + "/" \
                + wallpaper_dict["file"]
        return fullpath

    def set_wallpaper(self, wallpaper_dict: dict, feh_options: tuple=("--bg-fill",)) -> None:
        """
        Sets the wallpaper given in wallpaper_dict, according to the options in wallpaper_config,
        using feh.
        :param Dict wallpaper_dict: Wallpaper dictionary, with "file" containing the full path
        :param tuple feh_options: list of arguments to be supplied to feh
        """
        # feh is used, and not pywal.wallpaper.change() as the latter doesn't support feh arguments
        subprocess.run(["feh", wallpaper_dict["file"]] + list(feh_options), check=True)

    def get_theme(self, wallpaper_dict: dict, wallpaper_config: WallpaperConfig):
        """
        Get the theme name.

        :param Dict wallpaper_dict: The object describing a wallpaper
        :param WallpaperConfig wallpaper_config: wallpaper configuration
        :return: The name of the theme, compatible with `wal --theme`
        :rtype: str
        """
        if wallpaper_config.theme["use_static_theme"]:
            theme = wallpaper_config.theme["default_theme"]
        else:
            theme = wallpaper_dict["theme"] if "theme" in wallpaper_dict else None
        return theme

    def set_theme(self, wallpaper_path: str, theme: str=None):
        """
        Set the theme

        :param str wallpaper_path: Path to the wallpaper. This is the final path, and no prefix
        will be added, contrary to what may be store in the objects of the metadata files.
        :param str theme: Name of the theme, compatible with `wal --theme`.
        """
        if theme is not None:
            colors = pywal.theme.file(theme)
        else:
            # bug when for covers, see https://github.com/dylanaraps/pywal/issues/429
            # cache is not reset, so we have to do it by hand first - SHOULD BE FIXED
            # (require test)
            # The two lines below do the following : wal -c
            #scheme_dir = os.path.join(CACHE_DIR, "schemes")
            #shutil.rmtree(scheme_dir, ignore_errors=True)
            colors = pywal.colors.get(wallpaper_path)
        pywal.sequences.send(colors, vte_fix=True)
        pywal.export.every(colors)
        pywal.reload.env()
        # fixing powerline colors
        # subprocess.run([abspath(expanduser(wallpaper_config.theme["powerline_colours"]))],
        # check=True)
        


#! /usr/bin/python
"""
This is the main script that pyrrot should launch.

It is currently in the process of being cleaned and refactored.
"""

## TODO
#
# * If the song is paused (for more than x seconds), use a "normal" wallpaper
# * CONFIG_FILE override

from os.path import expanduser, dirname

from pyrrot_wallpaper.wallpaper_metadata import WallpaperMetadata
from pyrrot_wallpaper.config import SelectionMode, WallpaperConfig
from pyrrot_wallpaper.wallpaper_setter import WallpaperSetter

def main():
    wallpaper_config = WallpaperConfig(expanduser("~") + "/.config/pyrrot/pyrrot.config")

    if wallpaper_config.debug:
        print(f"Mode: {wallpaper_config.selection_mode}")

    wallpaperMetdata = WallpaperMetadata(
        expanduser(dirname(wallpaper_config.metadata_file)),
        expanduser(wallpaper_config.metadata_file)
        )

    """
    # TODO : use argparse or something like this
    flag_tags = False
    flag_pinfo = False
    flag_ptags = False
    for arg in sys.argv:
        if flag_tags:
            TAGS = arg.split(',')
            for i in range(len(TAGS)):
                TAGS[i] = TAGS[i].strip(' ')
            flag_tags = False
        if arg == "--tags":
            flag_tags = True
        if arg == "--pinfos":
            flag_pinfo = True
        if arg == "--ptags":
            flag_ptags = True
        if arg == "--albumart":
            USE_ALBUMART = True

    if flag_ptags:
        all_tags = print_all_tags(infos)
        all_tags.sort()
        for t in all_tags:
            print(t)
    elif flag_pinfo:
        print(TAGS)
        for pic in get_wallpapers_with_tags(infos):
            print(pic["name"])
    elif USE_ALBUMART:
        pic = {}
        set_wallpaper(pic);
    else:
        set_wallpaper(select_wallpaper(infos))
    """
    wallpaper_setter = WallpaperSetter()
    if wallpaper_config.selection_mode == SelectionMode.ALBUMART:
        pic = {}
    else:
        pic = wallpaperMetdata.select_single_wallpaper(wallpaper_config)
    pic["file"] = wallpaper_setter.get_wallpaper_full_path(pic, wallpaper_config)
    wallpaper_setter.set_wallpaper(pic)
    if wallpaper_config.theme["do_update_theme"]:
        wallpaper_setter.set_theme(pic["file"],
            theme=wallpaper_setter.get_theme(pic, wallpaper_config))


if __name__ == '__main__':
    sys.exit(main())
